// ignore_for_file: depend_on_referenced_packages, file only used in dev, dependencies are in dev_dependencies

import 'dart:async';

import 'package:api/annotations/authenticated/authenticated.dart';
import 'package:build/build.dart';
import 'package:source_gen/source_gen.dart' as g;
import 'package:analyzer/dart/element/element.dart';
import 'package:code_builder/code_builder.dart' as code;

const _authenticatedType = g.TypeChecker.fromRuntime(Authenticated);

List<MethodElement> getAnnotatedMethodsOrderBySourceOffset(
        ClassElement cls) =>
    <MethodElement>[
      ...cls.methods.where(_authenticatedType.hasAnnotationOfExact),
    ]..sort((a, b) => (a.nameOffset).compareTo(b.nameOffset));

code.Parameter getParameterFromParameterElement(ParameterElement element) {
  var isRequired = !element.isPositional && element.isRequired;
  var defaultTo = element.defaultValueCode != null ? code.Code(element.defaultValueCode!) : null;

  return code.Parameter((p) => p
    ..named = element.isNamed
    ..required = isRequired
    ..name = element.name
    ..defaultTo = defaultTo
    ..type = code.refer(element.type.getDisplayString())
  );
}

bool isParameterRequired(ParameterElement element) {
  return !element.isPositional && element.isRequired;
}

class AuthenticatedGenerator extends g.Generator {
  @override
  FutureOr<String?> generate(g.LibraryReader library, BuildStep buildStep) {
    for(var cls in library.classes){
      var className = cls.displayName;
      var classMethods = <code.Method>[];

      var annotatedMethods = getAnnotatedMethodsOrderBySourceOffset(cls);
      if (annotatedMethods.isNotEmpty) {
        log.info('found Authenticated annotations in $className');

        for (var method in annotatedMethods){
          var methodName = method.displayName;
          var returnType = method.returnType.getDisplayString();

          var methodBuilder = code.MethodBuilder()
            ..name = methodName
            ..annotations.add(code.CodeExpression(code.Code('override')))
            ..returns = code.refer(returnType);

          var requiredParameters = method.parameters.where((p) => !p.isNamed);
          var optionalParameters = method.parameters.where((p) => p.isNamed);

          methodBuilder
            ..requiredParameters.addAll(requiredParameters.map(getParameterFromParameterElement))
            ..optionalParameters.addAll(optionalParameters.map(getParameterFromParameterElement));

          var parameterForwardings = requiredParameters
            .map((p) => p.displayName)
            .toList()
            ..addAll(
              optionalParameters.map((p) => '${p.displayName}: ${p.displayName}')
            );
          var parameterForwarding = parameterForwardings.join(', ');

          var methodCode = <code.Code>[
            code.Code('print(\'test\');'),
          ];

          if (returnType == 'void') {
            methodCode.addAll([
              code.Code('super.$methodName($parameterForwarding);',),
            ]);
          } else {
            methodCode.addAll([
              code.Code('return super.$methodName($parameterForwarding);'),
            ]);
          }

          methodBuilder.body = code.Block.of(methodCode);

          classMethods.add(methodBuilder.build());
        }
      }

      if(classMethods.isNotEmpty) {
        return code.Library((lb) =>
          lb.body.add(
            code.Class((mb) => mb
              ..name = className.replaceFirst(RegExp('_'), '')
              ..methods.addAll(classMethods)
              ..extend = code.refer(className)
            )
          )
        )
          .accept(code.DartEmitter())
          .toString();
      }
    }

    return null;
  }
}

Builder authenticatedGeneratorFactory(BuilderOptions _) =>
    g.SharedPartBuilder([AuthenticatedGenerator()], 'authenticated_generator');
